#!/usr/bin/env python3
#
# SPDX-FileCopyrightText: 2025 Carlos M. Fonseca <cmfonsec@dei.uc.pt>
#
# SPDX-License-Identifier: Apache-2.0


from api.base import Neighbourhood, Move
import math
from api.utils import sparse_fisher_yates_iter


# ------------------------------- Neighbourhood ------------------------------

class TwoOptNeighbourhood(Neighbourhood):
    def moves(self, solution):
        assert self.problem == solution.problem
        # raise NotImplementedError("Not implemented yet")
        n = self.problem.n
        if solution.is_feasible():
            for ix in range(1, n-1):
                for jx in range(ix + 2, n + (ix != 1)):
                    yield TwoOptMove(self, ix, jx)


    def random_moves_without_replacement(self, solution):
        assert self.problem == solution.problem
        # The Fisher-Yates shuffle algorithm is used to generate a random
        n = self.problem.n
        if solution.is_feasible():
            for x in sparse_fisher_yates_iter(n*(n-3)//2):
                jx = (1 + math.isqrt(1+8*x))//2
                ix = x - jx*(jx-1)//2 + 1
                jx += 2
                # Handle special case
                if ix == 1 and jx == n:
                    ix = n - 2
                yield TwoOptMove(self, ix, jx)


# ----------------------------------- Move -----------------------------------

class TwoOptMove(Move):
    def __init__(self, neighbourhood, ix, jx):
        self.neighbourhood = neighbourhood
        # ix and jx are indices
        self.ix = ix
        self.jx = jx

    def __str__(self):
        return "%d:%d" % (self.i, self.j)

    def apply(self, solution):
        prob = solution.problem
        # Update tour length
        # Update solution
        # raise NotImplementedError("Not implemented yet")
        n, ix, jx = prob.n, self.ix, self.jx
        # Update tour length
        t = solution.tour
        solution.lb -= prob.dist[t[ix-1]][t[ix]] + prob.dist[t[jx-1]][t[jx % n]]
        solution.lb += prob.dist[t[ix-1]][t[jx-1]] + prob.dist[t[ix]][t[jx % n]]
        # Update solution
        solution.tour[ix:jx] = solution.tour[ix:jx][::-1]

    def objective_value_increment(self,  solution):
        prob = solution.problem
        # Tour length increment
        #raise NotImplementedError("Not implemented yet")
        n, ix, jx = prob.n, self.ix, self.jx
        # Tour length increment
        t = solution.tour
        incr = prob.dist[t[ix-1]][t[jx-1]] + prob.dist[t[ix]][t[jx % n]]
        incr -= prob.dist[t[ix-1]][t[ix]] + prob.dist[t[jx-1]][t[jx % n]]
        return incr

