#!/usr/bin/env python3
#
# SPDX-FileCopyrightText: 2025 Carlos M. Fonseca <cmfonsec@dei.uc.pt>
#
# SPDX-License-Identifier: Apache-2.0

import random
import matplotlib.pyplot as plt  

from api.neighborhood_constr_add import AddNeighbourhood
from api.neighborhood_local_2ex import TwoOptNeighbourhood


class Solution:
    def __init__(self, problem, tour, not_visited, lb):
        self.problem = problem
        self.tour = tour
        self.not_visited = not_visited
        self.lb = lb
        self.c_nbhood = None
        self.l_nbhood = None

    def __str__(self):
        return " ".join(map(str, self.tour))

    def copy(self):
        return self.__class__(self.problem,
            self.tour.copy(),
            self.not_visited.copy(),
            self.lb)

    @classmethod
    def empty_solution(cls, problem):
        return cls(problem, [0], set(range(1, problem.n)), 0)

    @classmethod
    def random_solution(cls, problem):
        c = list(range(1, problem.n))
        random.shuffle(c)
        c.insert(0, 0)
        obj = problem.dist[c[-1]][c[0]]
        for ix in range(1, problem.n):
            obj += problem.dist[c[ix-1]][c[ix]]
        return Solution(problem, c, set(), obj)

    def problem_instance(self):
        return self.problem

    def is_feasible(self):
        return len(self.not_visited) == 0

    def objective_value(self):
        if self.is_feasible():
            return self.lb

    def lower_bound(self):
        return self.lb

    def construction_neighbourhood(self):
        if self.c_nbhood is None:
            self.c_nbhood = AddNeighbourhood(self.problem)
        return self.c_nbhood

    def local_neighbourhood(self):
        if self.l_nbhood is None:
            self.l_nbhood = TwoOptNeighbourhood(self.problem)
        return self.l_nbhood
    

    def plot_tour(self, style='bo-', hilite='rs', title=''): 
        "Plot every city and link in the tour, and highlight the start city."
        scale = 1 + len(self.tour) ** 0.5 // 10
        plt.figure(figsize=((3 * scale, 2 * scale)))
        start = self.tour[0]        
        self.plot_segment([*self.tour, start], style)
        self.plot_segment([start], hilite) 
        plt.title(title)
        

    def plot_segment(self, segment: list, style='bo:'):
        "Plot every city and link in the segment."
        def Xs(cities) -> list[float]: "X coordinates"; return [self.problem.xy[c][0] for c in cities]
        def Ys(cities) -> list[float]: "Y coordinates"; return [self.problem.xy[c][1] for c in cities]

        plt.plot(Xs(segment), Ys(segment), style, linewidth=2/3, markersize=4, clip_on=False)
        plt.axis('scaled'); plt.axis('off')
